-- map.e: miscellaneous mapping routines
-- j.babor@gns.cri.nz
-- 98-07-30

global function scale_index(integer N,integer n)
    -- generate index sequence to remap evenly any sequence
    -- of length N to new length n
    sequence s
    integer i,j,nn,nN,NN,dx,dX,e
    s=repeat(0,n)
    if n=N then for k=1 to n do s[k]=k end for
    else
	nn=n*n
	nN=n*N
	NN=N*N
	dx=2*(NN-nN)
	dX=2*(nN-nn)
	e=nN-NN
	if n>N then
	    dx=-dx
	    dX=-dX
	    e=-e
	end if
	i=0
	j=0
	while j<n do
	    if e<0 then
		i=i+1
		e=e+dX
	    else
		j=j+1
		s[j]=i
		e=e-dx
	    end if
	end while
    end if
    return s
end function

global function inverse_index(integer m,integer n,atom ro)
    -- generate spacing index sequence (order of rows to be to be taken
    -- from the original sequence to generate remapped sequence) to remap
    -- rectangular image to a one of linearly variable width
    -- m... length of original sequence
    -- n... length of remapped sequence
    -- ro is ratio of lengths of the first and last row resp. of new image
    -- assumption: row spacing density is *inversely* proportional to row
    -- length
    sequence s
    atom dx,ddx,r,p,q,t
    s=repeat(0,n)
    r=m/n
    p=r/(ro+1)
    q=r*(ro-1)/(ro+1)/n
    t=p+q
    dx=2*p
    ddx=2*q
    for i=1 to n do
        s[i]=floor(t)+1
        dx=dx+ddx
        t=t+dx
    end for
    return s
end function

global function linear_spacing(integer l,integer i1, integer i2)
    -- generate integer sequence of length l with even distribution
    -- of values between i1 and i2 limits
    sequence s
    atom y,dy
    s=repeat(0,l)
    if i2<i1 then
        dy=(i1-i2+1)/l
        y=-dy/2
        for i=1 to l do
            s[i]=i1+floor(y)+1
            y=y-dy
        end for
    else
        dy=(i2-i1+1)/l
        y=dy/2
        for i=1 to l do
            s[i]=i1+floor(y)
            y=y+dy
        end for
    end if
    return s
end function

global function scale_sequence(sequence S,integer n)
    -- rescale sequence S to new length n
    sequence s
    integer i,j,N,nn,nN,NN,dx,dX,e
    N=length(S)
    if n=N then return S
    else
        s=repeat(0,n)
        nn=n*n
        nN=n*N
        NN=N*N
        dx=2*(NN-nN)
        dX=2*(nN-nn)
        e=nN-NN
        if n>N then
            dx=-dx
            dX=-dX
            e=-e
        end if
        i=0
        j=0
        while j<n do
            if e<0 then
                i=i+1
                e=e+dX
            else
                j=j+1
                s[j]=S[i]
                e=e-dx
            end if
        end while
        return s
    end if
end function

global function scale_image(sequence im,integer w,integer h)
    -- rescale *rectangular* sequence im to new width w and height h
    -- great ideas from Colin Taylor, further optimization by Rob Craig
    sequence r,s,t
    sequence ti, imri
    integer l
    l=length(im)
    r=scale_index(l,h)
    l=length(im[1])
    s=scale_index(l,w)
    t=repeat(repeat(0,w),h)
    for i=1 to h do
	ti = t[i]
	imri = im[r[i]]
	for j=1 to w do ti[j]=imri[s[j]] end for
	t[i] = ti
    end for
    return t
end function

global function map(sequence s,integer wt,integer wb,integer h)
    -- remap rectangular sequence s to new image of linearly
    -- variable width from wt at topo to wb at bottom
    sequence os,ws,ys
    integer wi,he
    he=length(s)
    wi=length(s[1])
    os=repeat({},h)
    ws=linear_spacing(h,wt,wb)
    ys=inverse_index(he,h,wt/wb)
    for i=1 to h do
        os[i]=scale_sequence(s[ys[i]],ws[i])
    end for
    return os
end function
