-- sky.ex: sky loop with perspective mapping
-- j.babor@gns.cri.nz
-- 98-07-30

without type_check
without warning

include get.e
include image.e
include graphics.e
include ememcopy.e          -- thanx Michael, marvelous...
include map.e

object junk
sequence p,sky,clouds,row_offset,index,iseq
atom ac,as,amap,amix,t,ts,acp,asp
integer A,fsf,jump_line,offset,w
integer frame_count,key,nc,ns,ncs

procedure initialize()
    integer wi
    tick_rate(100)          -- better timer resolution
    junk=set_up_emc()
    set_mask_color(2,200)
    clear_screen()
    text_color(14)
    position(2,2)
    puts(1,"Generating fractal clouds - please wait...")
    fsf=0                   -- full screen flag
    w=513                   -- length of buffer line
    A=#A3E80                -- 50th row of video memory
    jump_line=258           -- 2*127
    offset=132354           -- 2*127*513
    index=scale_index(257,320)
    row_offset=513*(inverse_index(100,100,10)-1)    -- of original image rows
    iseq={}
    wi=216
    for i=1 to 100 do
        iseq=append(iseq,scale_index(wi,320))
        wi=wi+2
    end for
end procedure

procedure allocate_memory()
    ac=allocate(228480)     -- clouds (2*257+200)*320
    as=allocate(228480)     -- sky
    amix=allocate(64000)    -- combine images: sky & clouds (200*320)
    amap=allocate(64000)    -- remapped sky & clouds
end procedure

function fractal(sequence s, integer depth, integer p, atom f,
    integer min, integer max)
    -- width w and height h of fractal image should be  n*(2^depth)+1
    -- where n is a positive integer
    -- depth ... number of iterations
    -- p ... initial maximum perturbation
    -- f ... fractional amplification/attenuation
    -- min, max ... total pertubation limits
    integer step,halfstep,p1,p2,w,h
    h=length(s)
    w=length(s[1])
    step=power(2,depth)
    while step>1 do
        halfstep=step/2
        p1=2*p+1
        p2=p+1
        for y=1 to h by step do
            for x=1 to w by step do
                s[y][x]=s[y][x]+rand(p1)-p2
            end for
            for x=1+halfstep to w-halfstep by step do
                s[y][x]=floor((s[y][x-halfstep]+s[y][x+halfstep])/2)
            end for
        end for
        for y=1+halfstep to h-halfstep by step do
            for x=1 to w by halfstep do
                s[y][x]=floor((s[y-halfstep][x]+s[y+halfstep][x])/2)
            end for
        end for
        step=halfstep
        p=floor(f*p)
    end while
    for i=1 to depth do
    end for
    for y=1 to h do
        for x=1 to w do
            if s[y][x]<min then s[y][x]=min
            elsif s[y][x]>max then s[y][x]=max end if
        end for
    end for
    return s
end function

procedure make_heavens()
    sequence s,sc,ss,sci,ssi
    if fsf then
        s=repeat(repeat(200,257),257)
        clouds=fractal(s,7,65,0.65,200,255)
        s=repeat(repeat(172,257),257)
        sky=fractal(s,7,70,0.6,128,255)
        -- widen fractal sequences from 257 to 320 pixels
        sc=repeat(0,320)
        ss=sc
        for i=1 to 257 do
            sci=clouds[i]
            ssi=sky[i]
            for j=1 to 320 do
                sc[j]=sci[index[j]]
                ss[j]=ssi[index[j]]
            end for
            clouds[i]=sc
            sky[i]=ss
        end for
    else
        s=repeat(repeat(200,513),129)
        clouds=fractal(s,7,65,0.65,200,255)
        s=repeat(repeat(172,513),129)
        sky=fractal(s,7,70,0.6,128,255)
    end if
end procedure

procedure set_buffers()
    atom x
    if fsf then
        -- clouds buffer
        x=ac
        for i=1 to 257 do
            poke(x,clouds[i])
            x=x+320
        end for
        for i=257 to 1 by -1 do
            poke(x,clouds[i])
            x=x+320
        end for
        -- overlap
        for i=1 to 200 do
            poke(x,clouds[i])
            x=x+320
        end for
        -- high sky buffer
        x=as
        for i=1 to 257 do
            poke(x,sky[i])
            x=x+320
        end for
        for i=257 to 1 by -1 do
            poke(x,sky[i])
            x=x+320
        end for
        for i=1 to 200 do
            poke(x,sky[i])
            x=x+320
        end for
    else
        -- clouds buffer
        x=ac
        for i=1 to 129 do
            poke(x,clouds[i])
            x=x+513
        end for
        for i=129 to 1 by -1 do
            poke(x,clouds[i])
            x=x+513
        end for
        -- overlap
        for i=1 to 100 do
            poke(x,clouds[i])
            x=x+513
        end for
        -- high sky buffer
        x=as
        for i=1 to 129 do
            poke(x,sky[i])
            x=x+513
        end for
        for i=129 to 1 by -1 do
            poke(x,sky[i])
            x=x+513
        end for
        for i=1 to 100 do
            poke(x,sky[i])
            x=x+513
        end for
    end if
end procedure

procedure set_palette()
    p=get_all_palette()     -- fetch current palette
    for i=0 to 31 do p[129+i]={32+i/8,32+i/8,48+i/16} end for
    for i=0 to 63 do p[161+i]={36+3*i/8,36+3*i/8,50+3*i/16} end for
    for i=0 to 31 do p[225+i]={60+i/8,60+i/8,62+i/16} end for
    all_palette(p)
end procedure

procedure show_text()
    text_color(9)
    position(3,13)
    puts(1,"Fractal Clouds")
    text_color(4)
    position(22,15)
    puts(1,"Esc to exit")
    position(23,6)
    puts(1,"Space to toggle to full screen")
    position(24,8)
    puts(1,"Any other key to alter sky")
end procedure

procedure frame(atom sa,atom ca)
    sequence s,is,si
    atom a,b
    integer w
    if fsf then
        mem_copy(amix,sa,64000)             -- copy high sky to virtual screen
        se_mem_copy(2,ca,amix,320,200,0,0)  -- copy clouds to virtual screen
        mem_copy(A,amix,64000)              -- copy to screen
    else
        mem_copy(amix,sa,51300)             -- copy high sky to virtual screen
        se_mem_copy(2,ca,amix,513,100,0,0)  -- copy clouds to virtual screen
        a=amap
        b=amix+100
        w=216
        si=repeat(0,320)
        for i=1 to 100 do
            s=peek({b+row_offset[i]-i,w})
            is=iseq[i]
            for j=1 to 320 do
                si[j]=s[is[j]]
            end for
            poke(a,si)
            w=w+2
            a=a+320
        end for
        mem_copy(A,amap,32000)              -- copy to screen
    end if
end procedure


-- main ------------------------------------------------------------------------

initialize()
allocate_memory()
make_heavens()
set_buffers()
junk=graphics_mode(19)
set_palette()
clear_screen()
show_text()


acp=ac+offset
asp=as
nc=0  ns=0  ncs=0
frame_count=0
t=time()
while 1 do
    ts=time()           -- frame start time
    frame(asp,acp)
    if nc>=jump_line then
        nc=0
        acp=ac+offset
    else
        nc=nc+1
        acp=acp-w
    end if
    ncs=ncs+1
    if ncs>=3 then
        ncs=0
        if ns>=jump_line then
            ns=0
            asp=as
        else
            ns=ns+1
            asp=asp+w
        end if
    end if
    frame_count=frame_count+1
    key=get_key()
    if key!=-1 then
        if key=27 then exit end if
        if key=32 then fsf=not fsf end if
        if fsf then
            A=#A0000            -- start of video memory
            jump_line=514       -- 2*257
            offset=164480       -- 2*257*320
            w=320               -- length of buffer line
        else
            A=#A3E80            -- 50th row of video memory
            jump_line=258       -- 2*127
            offset=132354       -- 2*127*513
            w=513               -- length of buffer line
        end if
        acp=ac+offset
        asp=as
        nc=0  ns=0  ncs=0
        make_heavens()
        set_buffers()
        if not fsf then
            clear_screen()
            show_text()
        end if
        t=time()
        frame_count=0
    end if
    while time()-ts<0.03 do end while   -- limit to 33 fps
end while
t=time()-t

junk=graphics_mode(-1)
printf(1,"%.2f fps\n", frame_count/t)
